import io
import re
import zipfile
from collections.abc import Mapping

import openpyxl
import xltpl.writerx as writerx
from docx import Document
from docxtpl import DocxTemplate
from jinja2.exceptions import TemplateSyntaxError
from mailmerge import MailMerge
from rest_framework import exceptions

from document_merge_service.api.data import django_file

from . import models
from .jinja import get_jinja_env, get_jinja_filters


class _MagicPlaceholder(str, Mapping):  # type: ignore
    def __new__(cls, parent=None, name=None):
        self = str.__new__(cls, name if name else "")
        self._parent = parent
        self._reports = parent._reports if parent else set()

        if self != "":
            self._reports.add(self)
        return self

    @property
    def reports(self):
        return list(self._reports)

    def __iter__(self):
        return (x for x in [_MagicPlaceholder(parent=self, name=f"{self}[]")])

    def __getitem__(self, idx):
        assert isinstance(idx, str)
        return _MagicPlaceholder(parent=self, name=f"{self}.{idx}".strip("."))

    def __getattr__(self, attr):
        return _MagicPlaceholder(parent=self, name=f"{self}.{attr}".strip("."))

    def __len__(self):
        return 2

    def __radd__(self, other):
        return str(self) + str(other)


class DocxValidator:
    def _validate_is_docx(self):
        try:
            Document(self.template)
        except (ValueError, zipfile.BadZipfile):
            raise exceptions.ParseError("not a valid docx file")
        finally:
            self.template.seek(0)

    def validate_template_syntax(self, available_placeholders=None):  # pragma: no cover
        raise NotImplementedError(
            "validate_template_syntax must be implemented in engine class"
        )

    def validate(self, available_placeholders=None, sample_data=None):
        self._validate_is_docx()
        self.validate_template_syntax(available_placeholders, sample_data)

    def validate_available_placeholders(
        self, used_placeholders, available_placeholders
    ):
        # We don't validate available_placeholders if it's not given
        if not available_placeholders:
            return

        available_placeholders = self._normalize_available_placeholders(
            available_placeholders
        )

        referenced_unavailable = "; ".join(
            sorted(set(used_placeholders) - set(available_placeholders))
        )
        if referenced_unavailable:
            raise exceptions.ValidationError(
                f"Template uses unavailable placeholders: {referenced_unavailable}"
            )

    def _normalize_available_placeholders(self, placeholders):
        available_placeholders = set(placeholders)
        # add all prefixes of placeholders, so users don't
        # have to add "foo" if they have "foo.bar" in the list
        for ph in placeholders:
            prefix = ""
            for word in ph.split("."):
                prefix = f"{prefix}.{word}" if prefix else word
                if prefix.endswith("[]"):
                    available_placeholders.add(prefix[:-2])
                available_placeholders.add(prefix)
        return available_placeholders


class DocxTemplateEngine(DocxValidator):
    def __init__(self, template):
        self.template = template

    def validate_template_syntax(self, available_placeholders=None, sample_data=None):

        try:
            doc = DocxTemplate(self.template)
            root = _MagicPlaceholder()
            env = get_jinja_env()
            ph = {
                name: root[name] for name in doc.get_undeclared_template_variables(env)
            }

            xml = doc.get_xml()
            xml = doc.patch_xml(xml)
            image_match = re.match(r".*{{\s?(\S*)\s?\|\s?image\(.*", xml)
            images = image_match.groups() if image_match else []
            for image in images:
                cleaned_image = image.strip('"').strip("'")
                ph[root[cleaned_image]] = django_file("black.png").file

            ph["_tpl"] = doc

            doc.render(ph, env)

            if sample_data:
                sample_data["_tpl"] = doc
                doc.render(sample_data, env)

            self.validate_available_placeholders(
                used_placeholders=root.reports,
                available_placeholders=available_placeholders,
            )

        except TemplateSyntaxError as exc:
            arg_str = ";".join(exc.args)
            raise exceptions.ValidationError(f"Syntax error in template: {arg_str}")

        finally:
            self.template.seek(0)

    def merge(self, data, buf):
        doc = DocxTemplate(self.template)
        data["_tpl"] = doc

        doc.render(data, get_jinja_env(), autoescape=True)
        doc.save(buf)
        return buf


class DocxMailmergeEngine(DocxValidator):
    def __init__(self, template):
        self.template = template

    def _get_placeholders(self, document):
        return [
            field.attrib.get("name")
            for part in document.parts.values()
            for field in part.findall("//MergeField")
        ]

    def validate_template_syntax(self, available_placeholders=None, sample_data=None):
        document = MailMerge(self.template)
        # syntax can't be invalid as it's validated by office
        # suites. However we need to have *some* placeholders
        self.template.seek(0)
        used_placeholders = self._get_placeholders(document)
        self.validate_available_placeholders(
            used_placeholders=used_placeholders,
            available_placeholders=available_placeholders,
        )

        if sample_data:
            buffer = io.BytesIO()
            self.merge(sample_data, buffer)
            self.template.seek(0)

        if not len(used_placeholders):
            raise exceptions.ValidationError("Template has no merge fields")

    def merge(self, data, buf):
        with MailMerge(self.template) as document:
            document.merge(**data)
            document.write(buf)
            return buf


_placeholder_match = re.compile(r"^\s*{{\s*([^{}]+)\s*}}\s*$")


class XlsxTemplateEngine:
    BUILTIN_VARS = [
        "tpl_name",
        "sheet_name",
        "[]",
        "sheet_name.decode",
    ]

    def __init__(self, template):
        self.template = template
        self.writer = None

    def validate_is_xlsx(self):
        try:
            openpyxl.load_workbook(self.template)
        except (ValueError, zipfile.BadZipfile):
            raise exceptions.ParseError("not a valid xlsx file")

    def validate(self, available_placeholders=None, sample_data=None):
        self.validate_is_xlsx()
        self.validate_template_syntax(available_placeholders, sample_data)

    def _expand_available_placeholders(self, ph_list):
        """Expand available placeholder list for (internal) correctness.

        If client gives "foo[].bar", we implicitly also allow "foo[]" and "foo":

        >>> self._expand_available_placeholders(["foo[].bar", "baz.boo"])
        ["foo[]", "foo[].bar", "baz.boo"]
        """
        out_list = []
        for ph in ph_list:
            pieces = ph.split(".")
            for offset in range(len(pieces)):
                prefixed = ".".join(pieces[:offset])
                out_list.append(prefixed)
                if prefixed.endswith("[]"):
                    out_list.append(prefixed[:-2])
            out_list.append(ph)
        return out_list

    def validate_template_syntax(self, available_placeholders=None, sample_data=None):
        # We cannot use jinja to validate because xltpl uses jinja's lexer directly
        magic = None
        if not sample_data:
            sample_data = magic = _MagicPlaceholder()
        buf = io.BytesIO()

        try:
            self.merge(sample_data, buf, is_test_merge=True)
        except TemplateSyntaxError as exc:
            arg_str = ";".join(exc.args)
            raise exceptions.ValidationError(f"Syntax error in template: {arg_str}")

        if available_placeholders and magic is not None:

            missing_set = (
                set(magic.reports)
                - set(self._expand_available_placeholders(available_placeholders))
                - set(self.BUILTIN_VARS)
            )
            if not missing_set:
                return

            missing = "; ".join(missing_set)
            raise exceptions.ValidationError(
                f"Placeholders used in template, but not available: {missing}"
            )

    def merge(self, data, buf, is_test_merge=False):
        self.writer = writer = writerx.BookWriter(self.template)
        self._current_data = data

        writer.jinja_env.filters.update(get_jinja_filters())
        if is_test_merge:
            writer.jinja_env.undefined = self._undefined_factory
        writer.jinja_env.globals.update(dir=dir, getattr=getattr)

        payloads = []
        sheets = writer.sheet_resource_map.sheet_state_list
        for sheet in sheets:
            new = dict(data)
            new["sheet_name"] = sheet.name
            new["tpl_name"] = sheet.name
            payloads.append(new)
        writer.render_book(payloads=payloads)
        writer.save(buf)
        return buf

    def _undefined_factory(self, name):
        # For test merges, we set a custom "undefined" factory that
        # doesn't really do undefined, but just fetches the right value
        # from our magic placeholder structure
        return self._current_data[name]


ENGINES = {
    models.Template.DOCX_TEMPLATE: DocxTemplateEngine,
    models.Template.DOCX_MAILMERGE: DocxMailmergeEngine,
    models.Template.XLSX_TEMPLATE: XlsxTemplateEngine,
}


def get_engine(engine, template):
    return ENGINES[engine](template)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.17 on 2019-01-03 10:24
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Template",
            fields=[
                ("slug", models.SlugField(primary_key=True, serialize=False)),
                ("description", models.TextField(default="")),
                ("template", models.FileField(max_length=1024, upload_to="")),
                (
                    "engine",
                    models.CharField(
                        choices=[
                            (
                                "docx-template",
                                "Docx Template engine (https://github.com/elapouya/python-docx-template)",
                            ),
                            (
                                "docx-mailmerge",
                                "Docx MailMerge engine (https://github.com/Bouke/docx-mailmerge)",
                            ),
                        ],
                        max_length=20,
                    ),
                ),
            ],
        )
    ]



# Generated by Django 3.2.16 on 2022-12-23 12:54

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("api", "0005_xlsx_template_engine"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="template",
            name="group",
        ),
    ]



from django.conf import settings
from rest_framework.pagination import PageNumberPagination


class APIPagination(PageNumberPagination):
    page_size_query_param = "page_size"
    max_page_size = settings.PAGINATION_MAX_PAGE_SIZE



import pytest
from django.urls import reverse
from rest_framework import status


@pytest.mark.parametrize(
    "query_params,expected",
    [
        ({"page_size": 10}, 10),
        ({"page_size": 120}, 110),  # max page size reached
        ({}, 100),  # default page size
    ],
)
def test_pagination(db, client, template_factory, query_params, expected, mocker):
    mocker.patch(
        "document_merge_service.api.pagination.APIPagination.max_page_size", 110
    )

    template_factory.create_batch(120)

    response = client.get(reverse("template-list"), data=query_params)

    assert response.status_code == status.HTTP_200_OK
    result = response.json()
    assert result["count"] == 120
    assert len(result["results"]) == expected



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots[
    'test_merge_expression[{{NAME and ", represents " + NAME}}-template_content1] 1'
] = """<w:body xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main" xmlns:o="urn:schemas-microsoft-com:office:office" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships" xmlns:v="urn:schemas-microsoft-com:vml" xmlns:w10="urn:schemas-microsoft-com:office:word" xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing" xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape" xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup" xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" xmlns:wp14="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing" xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml">
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve">Test</w:t>
    </w:r>
    <w:r>
      <w:rPr>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve">: , represents foo</w:t>
    </w:r>
  </w:p>
  <w:sectPr>
    <w:type w:val="nextPage"/>
    <w:pgSz w:w="11906" w:h="16838"/>
    <w:pgMar w:left="1134" w:right="1134" w:header="0" w:top="1134" w:footer="0" w:bottom="1134" w:gutter="0"/>
    <w:pgNumType w:fmt="decimal"/>
    <w:formProt w:val="false"/>
    <w:textDirection w:val="lrTb"/>
    <w:docGrid w:type="default" w:linePitch="240" w:charSpace="0"/>
  </w:sectPr>
</w:body>
"""

snapshots[
    'test_merge_expression[{{NAME and ", represents " + NAME}}-template_content2] 1'
] = """<w:body xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main" xmlns:o="urn:schemas-microsoft-com:office:office" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships" xmlns:v="urn:schemas-microsoft-com:vml" xmlns:w10="urn:schemas-microsoft-com:office:word" xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing" xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape" xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup" xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" xmlns:wp14="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing" xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml">
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve">Test</w:t>
    </w:r>
    <w:r>
      <w:rPr>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve">: </w:t>
    </w:r>
  </w:p>
  <w:sectPr>
    <w:type w:val="nextPage"/>
    <w:pgSz w:w="11906" w:h="16838"/>
    <w:pgMar w:left="1134" w:right="1134" w:header="0" w:top="1134" w:footer="0" w:bottom="1134" w:gutter="0"/>
    <w:pgNumType w:fmt="decimal"/>
    <w:formProt w:val="false"/>
    <w:textDirection w:val="lrTb"/>
    <w:docGrid w:type="default" w:linePitch="240" w:charSpace="0"/>
  </w:sectPr>
</w:body>
"""

snapshots[
    "test_merge_expression[{{blah}}-template_content0] 1"
] = """<w:body xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main" xmlns:o="urn:schemas-microsoft-com:office:office" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships" xmlns:v="urn:schemas-microsoft-com:vml" xmlns:w10="urn:schemas-microsoft-com:office:word" xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing" xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape" xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup" xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" xmlns:wp14="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing" xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml">
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve">Test</w:t>
    </w:r>
    <w:r>
      <w:rPr>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve">: blub</w:t>
    </w:r>
  </w:p>
  <w:sectPr>
    <w:type w:val="nextPage"/>
    <w:pgSz w:w="11906" w:h="16838"/>
    <w:pgMar w:left="1134" w:right="1134" w:header="0" w:top="1134" w:footer="0" w:bottom="1134" w:gutter="0"/>
    <w:pgNumType w:fmt="decimal"/>
    <w:formProt w:val="false"/>
    <w:textDirection w:val="lrTb"/>
    <w:docGrid w:type="default" w:linePitch="240" w:charSpace="0"/>
  </w:sectPr>
</w:body>
"""

snapshots[
    "test_merge_expression[{{escapeme}}-template_content3] 1"
] = """<w:body xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main" xmlns:o="urn:schemas-microsoft-com:office:office" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships" xmlns:v="urn:schemas-microsoft-com:vml" xmlns:w10="urn:schemas-microsoft-com:office:word" xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing" xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape" xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup" xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" xmlns:wp14="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing" xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml">
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve">Test</w:t>
    </w:r>
    <w:r>
      <w:rPr>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve">: &lt;&amp;&gt;</w:t>
    </w:r>
  </w:p>
  <w:sectPr>
    <w:type w:val="nextPage"/>
    <w:pgSz w:w="11906" w:h="16838"/>
    <w:pgMar w:left="1134" w:right="1134" w:header="0" w:top="1134" w:footer="0" w:bottom="1134" w:gutter="0"/>
    <w:pgNumType w:fmt="decimal"/>
    <w:formProt w:val="false"/>
    <w:textDirection w:val="lrTb"/>
    <w:docGrid w:type="default" w:linePitch="240" w:charSpace="0"/>
  </w:sectPr>
</w:body>
"""

snapshots["test_template_detail 1"] = {
    "available_placeholders": None,
    "description": """Article star very capital morning option. Interesting station story.
Where during teach country talk across drop. Central meeting anyone remember. There today material minute ago get.""",
    "disable_template_validation": False,
    "engine": "docx-mailmerge",
    "meta": {},
    "sample_data": None,
    "slug": "note-act-source",
    "template": None,
}

snapshots[
    "test_template_merge_docx[TestNameMailMerge-docx-mailmerge-template__template1] 1"
] = """<w:body xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main" xmlns:wpc="http://schemas.microsoft.com/office/word/2010/wordprocessingCanvas" xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" xmlns:o="urn:schemas-microsoft-com:office:office" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships" xmlns:m="http://schemas.openxmlformats.org/officeDocument/2006/math" xmlns:v="urn:schemas-microsoft-com:vml" xmlns:wp14="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing" xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing" xmlns:w10="urn:schemas-microsoft-com:office:word" xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml" xmlns:w15="http://schemas.microsoft.com/office/word/2012/wordml" xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup" xmlns:wpi="http://schemas.microsoft.com/office/word/2010/wordprocessingInk" xmlns:wne="http://schemas.microsoft.com/office/word/2006/wordml" xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape">
  <w:p w:rsidR="0083709B" w:rsidRPr="0083709B" w:rsidRDefault="0083709B">
    <w:pPr>
      <w:rPr>
        <w:lang w:val="en-US"/>
      </w:rPr>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:lang w:val="en-US"/>
      </w:rPr>
      <w:t xml:space="preserve">Test: </w:t>
    </w:r>
    <w:r>
      <w:rPr>
        <w:lang w:val="en-US"/>
      </w:rPr>
      <w:t>Test input</w:t>
    </w:r>
    <w:bookmarkStart w:id="0" w:name="_GoBack"/>
    <w:bookmarkEnd w:id="0"/>
  </w:p>
  <w:sectPr w:rsidR="0083709B" w:rsidRPr="0083709B">
    <w:pgSz w:w="11906" w:h="16838"/>
    <w:pgMar w:top="1417" w:right="1417" w:bottom="1134" w:left="1417" w:header="708" w:footer="708" w:gutter="0"/>
    <w:cols w:space="708"/>
    <w:docGrid w:linePitch="360"/>
  </w:sectPr>
</w:body>
"""

snapshots[
    "test_template_merge_docx[TestNameTemplate-docx-template-template__template0] 1"
] = """<w:body xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main" xmlns:o="urn:schemas-microsoft-com:office:office" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships" xmlns:v="urn:schemas-microsoft-com:vml" xmlns:w10="urn:schemas-microsoft-com:office:word" xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing" xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape" xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup" xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" xmlns:wp14="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing" xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml">
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve">Test</w:t>
    </w:r>
    <w:r>
      <w:rPr>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve">: Test input</w:t>
    </w:r>
  </w:p>
  <w:sectPr>
    <w:type w:val="nextPage"/>
    <w:pgSz w:w="11906" w:h="16838"/>
    <w:pgMar w:left="1134" w:right="1134" w:header="0" w:top="1134" w:footer="0" w:bottom="1134" w:gutter="0"/>
    <w:pgNumType w:fmt="decimal"/>
    <w:formProt w:val="false"/>
    <w:textDirection w:val="lrTb"/>
    <w:docGrid w:type="default" w:linePitch="240" w:charSpace="0"/>
  </w:sectPr>
</w:body>
"""

snapshots[
    "test_template_merge_jinja_extensions_docx[docx-template-template__template0] 1"
] = """<w:body xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main" xmlns:o="urn:schemas-microsoft-com:office:office" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships" xmlns:v="urn:schemas-microsoft-com:vml" xmlns:w10="urn:schemas-microsoft-com:office:word" xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing" xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape" xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup" xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" xmlns:wp14="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing" xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml">
  <w:p>
    <w:pPr>
      <w:pStyle w:val="PreformattedText"/>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="3E4349"/>
        <w:spacing w:val="0"/>
      </w:rPr>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="B11414"/>
        <w:spacing w:val="0"/>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve"/>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="PreformattedText"/>
      <w:widowControl/>
      <w:shd w:val="clear" w:fill="EEEEEE"/>
      <w:spacing w:before="225" w:after="225"/>
      <w:ind w:left="0" w:right="0" w:hanging="0"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="B11414"/>
        <w:spacing w:val="0"/>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve"/>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="PreformattedText"/>
      <w:widowControl/>
      <w:shd w:val="clear" w:fill="EEEEEE"/>
      <w:spacing w:before="225" w:after="225"/>
      <w:ind w:left="0" w:right="0" w:hanging="0"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="B11414"/>
        <w:spacing w:val="0"/>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve">    </w:t>
    </w:r>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="B11414"/>
        <w:spacing w:val="0"/>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve"/>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="PreformattedText"/>
      <w:widowControl/>
      <w:shd w:val="clear" w:fill="EEEEEE"/>
      <w:spacing w:before="225" w:after="225"/>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:i w:val="false"/>
        <w:color w:val="B11414"/>
      </w:rPr>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="3E4349"/>
        <w:spacing w:val="0"/>
      </w:rPr>
      <w:t xml:space="preserve">    </w:t>
    </w:r>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="B11414"/>
        <w:spacing w:val="0"/>
      </w:rPr>
      <w:t xml:space="preserve">1</w:t>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="PreformattedText"/>
      <w:widowControl/>
      <w:shd w:val="clear" w:fill="EEEEEE"/>
      <w:spacing w:before="225" w:after="225"/>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="3E4349"/>
        <w:spacing w:val="0"/>
      </w:rPr>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="B11414"/>
        <w:spacing w:val="0"/>
      </w:rPr>
      <w:t xml:space="preserve"/>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="PreformattedText"/>
      <w:widowControl/>
      <w:shd w:val="clear" w:fill="EEEEEE"/>
      <w:spacing w:before="225" w:after="225"/>
      <w:ind w:left="0" w:right="0" w:hanging="0"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="B11414"/>
        <w:spacing w:val="0"/>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve">    </w:t>
    </w:r>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="B11414"/>
        <w:spacing w:val="0"/>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve"/>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="PreformattedText"/>
      <w:widowControl/>
      <w:shd w:val="clear" w:fill="EEEEEE"/>
      <w:spacing w:before="225" w:after="225"/>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:i w:val="false"/>
        <w:color w:val="B11414"/>
      </w:rPr>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="3E4349"/>
        <w:spacing w:val="0"/>
      </w:rPr>
      <w:t xml:space="preserve">    </w:t>
    </w:r>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="B11414"/>
        <w:spacing w:val="0"/>
      </w:rPr>
      <w:t xml:space="preserve">2</w:t>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="PreformattedText"/>
      <w:widowControl/>
      <w:shd w:val="clear" w:fill="EEEEEE"/>
      <w:spacing w:before="225" w:after="225"/>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="3E4349"/>
        <w:spacing w:val="0"/>
      </w:rPr>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="B11414"/>
        <w:spacing w:val="0"/>
      </w:rPr>
      <w:t xml:space="preserve"/>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="PreformattedText"/>
      <w:widowControl/>
      <w:shd w:val="clear" w:fill="EEEEEE"/>
      <w:spacing w:before="225" w:after="225"/>
      <w:ind w:left="0" w:right="0" w:hanging="0"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="B11414"/>
        <w:spacing w:val="0"/>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve">    </w:t>
    </w:r>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="B11414"/>
        <w:spacing w:val="0"/>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve"/>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="PreformattedText"/>
      <w:widowControl/>
      <w:shd w:val="clear" w:fill="EEEEEE"/>
      <w:spacing w:before="225" w:after="225"/>
      <w:ind w:left="0" w:right="0" w:hanging="0"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="B11414"/>
        <w:spacing w:val="0"/>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve">        </w:t>
    </w:r>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace" w:hAnsi="Consolas;Menlo;Deja Vu Sans Mono;Bitstream Vera Sans Mono;monospace"/>
        <w:b w:val="false"/>
        <w:i w:val="false"/>
        <w:caps w:val="false"/>
        <w:smallCaps w:val="false"/>
        <w:color w:val="B11414"/>
        <w:spacing w:val="0"/>
        <w:lang w:val="de-CH" w:eastAsia="zh-CN" w:bidi="hi-IN"/>
      </w:rPr>
      <w:t xml:space="preserve"/>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr/>
    </w:r>
  </w:p>
  <w:sectPr>
    <w:type w:val="nextPage"/>
    <w:pgSz w:w="11906" w:h="16838"/>
    <w:pgMar w:left="1134" w:right="1134" w:header="0" w:top="1134" w:footer="0" w:bottom="1134" w:gutter="0"/>
    <w:pgNumType w:fmt="decimal"/>
    <w:formProt w:val="false"/>
    <w:textDirection w:val="lrTb"/>
    <w:docGrid w:type="default" w:linePitch="240" w:charSpace="0"/>
  </w:sectPr>
</w:body>
"""

snapshots[
    "test_template_merge_jinja_filters_docx[docx-template-template__template0-False-False-200] 1"
] = """<w:body xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main" xmlns:o="urn:schemas-microsoft-com:office:office" xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships" xmlns:v="urn:schemas-microsoft-com:vml" xmlns:w10="urn:schemas-microsoft-com:office:word" xmlns:wp="http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing" xmlns:wps="http://schemas.microsoft.com/office/word/2010/wordprocessingShape" xmlns:wpg="http://schemas.microsoft.com/office/word/2010/wordprocessingGroup" xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" xmlns:wp14="http://schemas.microsoft.com/office/word/2010/wordprocessingDrawing" xmlns:w14="http://schemas.microsoft.com/office/word/2010/wordml">
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="DejaVu Sans" w:hAnsi="DejaVu Sans"/>
      </w:rPr>
      <w:t xml:space="preserve">15.09.1984</w:t>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="DejaVu Sans" w:hAnsi="DejaVu Sans"/>
      </w:rPr>
      <w:t xml:space="preserve">1984</w:t>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="DejaVu Sans" w:hAnsi="DejaVu Sans"/>
      </w:rPr>
      <w:t xml:space="preserve">23:24</w:t>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="DejaVu Sans" w:hAnsi="DejaVu Sans"/>
      </w:rPr>
      <w:t xml:space="preserve">15.09.1984, 23:23:00</w:t>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="DejaVu Sans" w:hAnsi="DejaVu Sans"/>
      </w:rPr>
      <w:t xml:space="preserve">23:23</w:t>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="DejaVu Sans" w:hAnsi="DejaVu Sans"/>
      </w:rPr>
      <w:t xml:space="preserve"/>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="DejaVu Sans" w:hAnsi="DejaVu Sans"/>
      </w:rPr>
      <w:t xml:space="preserve"/>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="DejaVu Sans" w:hAnsi="DejaVu Sans"/>
      </w:rPr>
      <w:t xml:space="preserve">something</w:t>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="DejaVu Sans" w:hAnsi="DejaVu Sans"/>
      </w:rPr>
      <w:t xml:space="preserve">This is</w:t>
      <w:br/>
      <w:t xml:space="preserve">a test.</w:t>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr/>
    </w:pPr>
    <w:r>
      <w:rPr>
        <w:rFonts w:ascii="DejaVu Sans" w:hAnsi="DejaVu Sans"/>
      </w:rPr>
      <w:t xml:space="preserve"/>
    </w:r>
    <w:r>
      <w:drawing>
        <wp:inline xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/main" xmlns:pic="http://schemas.openxmlformats.org/drawingml/2006/picture">
          <wp:extent cx="1800000" cy="1800000"/>
          <wp:docPr id="1001" name="Picture 1"/>
          <wp:cNvGraphicFramePr>
            <a:graphicFrameLocks noChangeAspect="1"/>
          </wp:cNvGraphicFramePr>
          <a:graphic>
            <a:graphicData uri="http://schemas.openxmlformats.org/drawingml/2006/picture">
              <pic:pic>
                <pic:nvPicPr>
                  <pic:cNvPr id="0" name="image.png"/>
                  <pic:cNvPicPr/>
                </pic:nvPicPr>
                <pic:blipFill>
                  <a:blip r:embed="rId5"/>
                  <a:stretch>
                    <a:fillRect/>
                  </a:stretch>
                </pic:blipFill>
                <pic:spPr>
                  <a:xfrm>
                    <a:off x="0" y="0"/>
                    <a:ext cx="1800000" cy="1800000"/>
                  </a:xfrm>
                  <a:prstGeom prst="rect"/>
                </pic:spPr>
              </pic:pic>
            </a:graphicData>
          </a:graphic>
        </wp:inline>
      </w:drawing>
    </w:r>
    <w:r>
      <w:t xml:space="preserve"/>
    </w:r>
  </w:p>
  <w:p>
    <w:pPr>
      <w:pStyle w:val="Normal"/>
      <w:rPr>
        <w:rFonts w:ascii="DejaVu Sans" w:hAnsi="DejaVu Sans"/>
      </w:rPr>
    </w:pPr>
    <w:r>
      <w:rPr/>
    </w:r>
  </w:p>
  <w:sectPr>
    <w:type w:val="nextPage"/>
    <w:pgSz w:w="11906" w:h="16838"/>
    <w:pgMar w:left="1134" w:right="1134" w:header="0" w:top="1134" w:footer="0" w:bottom="1134" w:gutter="0"/>
    <w:pgNumType w:fmt="decimal"/>
    <w:formProt w:val="false"/>
    <w:textDirection w:val="lrTb"/>
    <w:docGrid w:type="default" w:linePitch="240" w:charSpace="0"/>
  </w:sectPr>
</w:body>
"""



import json

import pytest
from django.urls import reverse
from rest_framework.status import HTTP_200_OK, HTTP_400_BAD_REQUEST


@pytest.mark.parametrize(
    "value,status_code",
    [
        (json.dumps([{"key": "foo", "value": "bar"}]), HTTP_200_OK),
        (json.dumps([{"key": "int", "value": 5, "lookup": "gt"}]), HTTP_200_OK),
        (
            json.dumps(
                [{"key": "foo", "value": "bar"}, {"key": "baz", "value": "bla"}]
            ),
            HTTP_200_OK,
        ),
        (
            json.dumps([{"key": "foo", "value": "bar", "lookup": "asdfgh"}]),
            HTTP_400_BAD_REQUEST,
        ),
        (json.dumps([{"key": "foo"}]), HTTP_400_BAD_REQUEST),
        (json.dumps({"key": "foo"}), HTTP_400_BAD_REQUEST),
        ("foo", HTTP_400_BAD_REQUEST),
        ("[{foo, no json)", HTTP_400_BAD_REQUEST),
    ],
)
def test_json_value_filter(db, template_factory, admin_client, value, status_code):
    doc = template_factory(meta={"foo": "bar", "baz": "bla", "int": 23})
    template_factory(meta={"foo": "baz"})
    template_factory()
    url = reverse("template-list")
    resp = admin_client.get(url, {"meta": value})
    assert resp.status_code == status_code
    if status_code == HTTP_200_OK:
        result = resp.json()
        assert len(result["results"]) == 1
        assert result["results"][0]["slug"] == str(doc.pk)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-01-08 13:09
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("api", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="template",
            name="group",
            field=models.CharField(
                blank=True, db_index=True, max_length=255, null=True
            ),
        )
    ]



import io
import json
import os

import openpyxl
import pytest
from django.urls import reverse
from docx import Document
from lxml import etree
from rest_framework import status

from document_merge_service.api.data import django_file

from .. import models, serializers


@pytest.mark.parametrize("template__description", ["test description"])
@pytest.mark.parametrize(
    "query_params,size",
    [
        ({"description__icontains": "test"}, 1),
        ({"description__search": "test"}, 1),
        ({"description__icontains": "unknown"}, 0),
        ({"description__search": "unknown"}, 0),
    ],
)
def test_template_list_query_params(db, admin_client, template, size, query_params):
    url = reverse("template-list")

    response = admin_client.get(url, data=query_params)
    assert response.status_code == status.HTTP_200_OK
    assert response.json()["count"] == size


def test_template_detail(db, client, template, snapshot):
    url = reverse("template-detail", args=[template.pk])

    response = client.get(url)
    assert response.status_code == status.HTTP_200_OK
    snapshot.assert_match(response.json())


def test_template_download(db, client, template):
    file = django_file("docx-template-syntax.docx")
    template.template.save(os.path.basename(file.name), file)
    template.save()

    url = reverse("template-detail", args=[template.pk])
    response = client.get(url)

    data = response.json()

    download_url = data["template"]

    template_resp = client.get(download_url)

    file.seek(0)
    assert file.read() == template_resp.content


def test_template_list_with_file(db, client, template):
    file = django_file("docx-template-syntax.docx")
    template.template.save(os.path.basename(file.name), file)
    template.save()

    url = reverse("template-list")
    response = client.get(url)

    assert response.json()["results"][0]["template"] is not None
    assert response.status_code == status.HTTP_200_OK


def test_template_download_url(db, client, template):
    file = django_file("docx-template-syntax.docx")
    template.template.save(os.path.basename(file.name), file)
    template.save()

    serializer = serializers.TemplateSerializer(template)
    field = serializer.fields["template"]
    assert (
        field.to_representation(template)
        == f"/api/v1/template-download/{template.slug}"
    )


@pytest.mark.parametrize(
    "template_name,engine,status_code,require_authentication,authenticated",
    [
        (
            "xlsx-template.xlsx",
            models.Template.XLSX_TEMPLATE,
            status.HTTP_201_CREATED,
            False,
            False,
        ),
        (
            "xlsx-template.xlsx",
            models.Template.XLSX_TEMPLATE,
            status.HTTP_201_CREATED,
            True,
            True,
        ),
        (
            "xlsx-template.xlsx",
            models.Template.XLSX_TEMPLATE,
            status.HTTP_401_UNAUTHORIZED,
            True,
            False,
        ),
        (
            "xlsx-template.xlsx",
            models.Template.XLSX_TEMPLATE,
            status.HTTP_201_CREATED,
            True,
            True,
        ),
        (
            "docx-template.docx",
            models.Template.DOCX_TEMPLATE,
            status.HTTP_201_CREATED,
            False,
            False,
        ),
        (
            "docx-template.docx",
            models.Template.DOCX_TEMPLATE,
            status.HTTP_201_CREATED,
            True,
            True,
        ),
        (
            "docx-template.docx",
            models.Template.DOCX_TEMPLATE,
            status.HTTP_401_UNAUTHORIZED,
            True,
            False,
        ),
        (
            "docx-template.docx",
            models.Template.DOCX_TEMPLATE,
            status.HTTP_201_CREATED,
            True,
            True,
        ),
        (
            "docx-mailmerge.docx",
            models.Template.DOCX_MAILMERGE,
            status.HTTP_201_CREATED,
            True,
            True,
        ),
        (
            "docx-mailmerge-syntax.docx",
            models.Template.DOCX_MAILMERGE,
            status.HTTP_400_BAD_REQUEST,
            True,
            True,
        ),
        (
            "docx-template-syntax.docx",
            models.Template.DOCX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
            True,
            True,
        ),
        (
            "test.txt",
            models.Template.DOCX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
            False,
            False,
        ),
    ],
)
def test_template_create(
    db,
    client,
    admin_client,
    engine,
    template_name,
    status_code,
    require_authentication,
    settings,
    authenticated,
):
    if authenticated:
        client = admin_client

    settings.REQUIRE_AUTHENTICATION = require_authentication

    url = reverse("template-list")

    template_file = django_file(template_name)
    data = {"slug": "test-slug", "template": template_file.file, "engine": engine}
    response = client.post(url, data=data, format="multipart")

    assert response.status_code == status_code

    if status_code == status.HTTP_201_CREATED:
        data = response.json()
        template_link = data["template"]
        response = client.get(template_link)
        assert response.status_code == status.HTTP_200_OK
        file_ = io.BytesIO(response.content)
        if engine == "xlsx-template":
            openpyxl.load_workbook(file_)
        else:
            Document(file_)


@pytest.mark.parametrize(
    "status_code, disable_validation",
    [
        (
            status.HTTP_400_BAD_REQUEST,
            "false",
        ),
        (
            status.HTTP_400_BAD_REQUEST,
            "",
        ),
        (
            status.HTTP_201_CREATED,
            "true",
        ),
    ],
)
def test_disable_validation(
    db,
    status_code,
    admin_client,
    settings,
    disable_validation,
):
    settings.REQUIRE_AUTHENTICATION = False
    url = reverse("template-list")

    template_file = django_file("docx-template-syntax.docx")
    data = {
        "slug": "test-slug",
        "template": template_file.file,
        "engine": models.Template.DOCX_TEMPLATE,
    }
    if disable_validation:
        data["disable_template_validation"] = disable_validation

    response = admin_client.post(url, data=data, format="multipart")
    assert response.status_code == status_code

    if status_code == status.HTTP_201_CREATED:
        data = response.json()
        template_link = data["template"]
        response = admin_client.get(template_link)
        assert response.status_code == status.HTTP_200_OK
        Document(io.BytesIO(response.content))


@pytest.mark.parametrize(
    "template_name,available_placeholders,sample_data,files,expect_missing_placeholders,engine,status_code",
    [
        (
            "docx-template-placeholdercheck.docx",
            ["foo", "bar", "baz"],
            None,
            [],
            [
                "bar.some_attr",
                "black.png",
                "list",
                "list[]",
                "list[].attribute",
            ],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
        ),
        (
            "docx-template-placeholdercheck.docx",
            [
                "foo",
                "bar",
                "baz",
                "bar.some_attr",
                "list[].attribute",
                "black.png",
            ],
            None,
            [],
            [],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_201_CREATED,
        ),
        (
            "docx-template-placeholdercheck.docx",
            [
                "foo",
                "bar",
                "baz",
                "bar.some_attr",
                "list[].attribute",
            ],
            None,
            [],
            ["black.png"],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
        ),
        (
            "docx-template-placeholdercheck.docx",
            None,
            {
                "foo": "hello",
                "bar": {
                    "some_attr": True,
                    "list": [{"attribute": "value"}, {"attribute": "value2"}],
                },
                "baz": "1234",
                "list": [{"attribute": "value"}],
            },
            [django_file("black.png").file],
            [],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_201_CREATED,
        ),
        (
            "docx-template-placeholdercheck.docx",
            None,
            {},
            [django_file("black.png").file],
            [],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
        ),
        (
            "docx-template-placeholdercheck.docx",
            None,
            {},
            [],
            [],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_201_CREATED,
        ),
        (
            "docx-template-placeholdercheck.docx",
            None,
            {
                "foo": "hello",
                "bar": {
                    "some_attr": True,
                    "list": [{"attribute": "value"}, {"attribute": "value2"}],
                },
                "baz": "1234",
                "list": [{"attribute": "value"}],
            },
            [],
            ["black.png"],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
        ),
        (
            "docx-template-placeholdercheck.docx",
            None,
            {
                "foo": "hello",
                "bar": {
                    "some_attr": True,
                    "list": [{"attribute": "value"}, {"attribute": "value2"}],
                },
            },
            [django_file("black.png").file],
            ["baz", "list", "list[]", "list[].attribute"],
            models.Template.DOCX_TEMPLATE,
            status.HTTP_400_BAD_REQUEST,
        ),
        (
            "docx-mailmerge.docx",
            None,
            {
                "foo": "hello",
                "bar": {
                    "some_attr": True,
                    "list": [{"attribute": "value"}, {"attribute": "value2"}],
                },
            },
            [],
            ["test"],
            models.Template.DOCX_MAILMERGE,
            status.HTTP_400_BAD_REQUEST,
        ),
        (
            "docx-mailmerge.docx",
            None,
            {
                "foo": "hello",
                "bar": {
                    "some_attr": True,
                    "list": [{"attribute": "value"}, {"attribute": "value2"}],
                },
            },
            [],
            ["test"],
            models.Template.DOCX_MAILMERGE,
            status.HTTP_400_BAD_REQUEST,
        ),
        (
            "docx-mailmerge.docx",
            None,
            {"test": "hello"},
            [],
            [],
            models.Template.DOCX_MAILMERGE,
            status.HTTP_201_CREATED,
        ),
        (
            "docx-mailmerge.docx",
            ["test", "blah"],
            {"test": "hello"},
            [],
            [],
            models.Template.DOCX_MAILMERGE,
            status.HTTP_400_BAD_REQUEST,
        ),
        (
            "docx-mailmerge.docx",
            [],
            {"test": "hello"},
            [django_file("black.png").file],
            [],
            models.Template.DOCX_MAILMERGE,
            status.HTTP_400_BAD_REQUEST,
        ),
    ],
)
def test_template_create_with_available_placeholders(
    db,
    admin_client,
    engine,
    template_name,
    available_placeholders,
    sample_data,
    files,
    status_code,
    settings,
    expect_missing_placeholders,
):

    settings.DOCXTEMPLATE_JINJA_EXTENSIONS = ["jinja2.ext.loopcontrols"]
    url = reverse("template-list")

    template_file = django_file(template_name)
    data = {
        "slug": "test-slug",
        "template": template_file.file,
        "files": files,
        "engine": engine,
    }
    if sample_data:
        data["sample_data"] = json.dumps(sample_data)
    if available_placeholders:
        data["available_placeholders"] = available_placeholders

    response = admin_client.post(url, data=data, format="multipart")
    assert response.status_code == status_code, response.json()

    if status_code == status.HTTP_400_BAD_REQUEST:
        resp = response.json()
        expect_missing_str = "; ".join(expect_missing_placeholders)

        if sample_data and available_placeholders:
            # validation only allows one of these two params
            assert (
                resp["non_field_errors"][0]
                == "Only one of available_placeholders and sample_data is allowed"
            )
        elif engine == models.Template.DOCX_MAILMERGE and files:
            assert (
                resp["non_field_errors"][0]
                == 'Files are only accepted with the "docx-template" engine'
            )
        elif not sample_data and files:
            assert (
                resp["non_field_errors"][0]
                == "Files are only accepted when also providing sample_data"
            )
        else:
            # we expect some missing placeholders
            assert resp["non_field_errors"][0] in [
                f"Template uses unavailable placeholders: {expect_missing_str}",
                f'No file for image "{expect_missing_str}" provided!',
            ]

    if status_code == status.HTTP_201_CREATED:
        data = response.json()
        template_link = data["template"]
        response = admin_client.get(template_link)
        assert response.status_code == status.HTTP_200_OK
        Document(io.BytesIO(response.content))


@pytest.mark.parametrize(
    "template__engine,template__template",
    [(models.Template.DOCX_TEMPLATE, django_file("docx-template-filters.docx"))],
)
@pytest.mark.parametrize(
    "template_name,status_code",
    [
        ("docx-template.docx", status.HTTP_200_OK),
        ("test.txt", status.HTTP_400_BAD_REQUEST),
    ],
)
def test_template_update(db, client, template, template_name, status_code):
    url = reverse("template-detail", args=[template.pk])

    template_file = django_file(template_name)
    data = {"description": "Test description", "template": template_file.file}
    response = client.patch(url, data=data, format="multipart")
    assert response.status_code == status_code

    if status_code == status.HTTP_200_OK:
        assert os.path.isfile(template.template.path) is False
        template.refresh_from_db()
        assert template.description == "Test description"


@pytest.mark.parametrize("template__template", [django_file("docx-template.docx")])
def test_template_destroy(db, client, template):
    url = reverse("template-detail", args=[template.pk])

    response = client.delete(url)
    assert response.status_code == status.HTTP_204_NO_CONTENT
    assert os.path.isfile(template.template.path) is False


@pytest.mark.parametrize(
    "template__slug,template__engine,template__template",
    [
        (
            "TestNameTemplate",
            models.Template.DOCX_TEMPLATE,
            django_file("docx-template.docx"),
        ),
        (
            "TestNameMailMerge",
            models.Template.DOCX_MAILMERGE,
            django_file("docx-mailmerge.docx"),
        ),
    ],
)
def test_template_merge_docx(
    db, client, mock_filefield_name_validation, template, snapshot
):
    url = reverse("template-merge", args=[template.pk])

    response = client.post(url, data={"data": {"test": "Test input"}}, format="json")
    assert response.status_code == status.HTTP_200_OK
    assert (
        response.get("content-type")
        == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    )

    docx = Document(io.BytesIO(response.content))
    xml = etree.tostring(docx._element.body, encoding="unicode", pretty_print=True)
    try:
        snapshot.assert_match(xml)
    except AssertionError:  # pragma: no cover
        with open(f"/tmp/{template.slug}.docx", "wb") as output:
            output.write(response.content)
        print("Template output changed. Check file at %s" % output.name)
        raise


@pytest.mark.parametrize(
    "placeholder,template_content",
    [
        ("{{blah}}", {"blah": "blub"}),
        (
            '{{NAME and ", represents " + NAME}}',
            {"NAME": "foo"},
        ),
        (
            '{{NAME and ", represents " + NAME}}',
            {"NAME": ""},
        ),
        # passed data should be escaped
        ("{{escapeme}}", {"escapeme": "<&>"}),
    ],
)
def test_merge_expression(
    docx_template_with_placeholder, client, snapshot, placeholder, template_content
):
    """Test evaluation of some custom expressions.

    Use this test to try out expressions without creating a new docx template for each
    variant.
    """
    template = docx_template_with_placeholder(placeholder)

    url = reverse("template-merge", args=[template.pk])

    response = client.post(url, data={"data": template_content}, format="json")
    assert response.status_code == status.HTTP_200_OK

    docx = Document(io.BytesIO(response.content))
    xml = etree.tostring(docx._element.body, encoding="unicode", pretty_print=True)
    try:
        snapshot.assert_match(xml)
    except AssertionError:  # pragma: no cover
        with open(f"/tmp/{template.slug}.docx", "wb") as output:
            output.write(response.content)
        print("Template output changed. Check file at %s" % output.name)
        raise


@pytest.mark.parametrize(
    "placeholder,template_content",
    [
        ("{{blah}}", {"blah": "blub"}),
        (
            '{{NAME and ", represents " + NAME}}',
            {"NAME": "foo"},
        ),
        (
            '{{NAME and ", represents " + NAME}}',
            {"NAME": ""},
        ),
    ],
)
@pytest.mark.parametrize(
    "template__engine",
    [models.Template.DOCX_TEMPLATE],
)
def test_validate_expression(
    docx_template_with_placeholder, client, placeholder, template_content
):
    """Test validation of templates with custom expressions."""
    template = docx_template_with_placeholder(placeholder)

    serializer = serializers.TemplateSerializer()
    serializer.instance = template

    serializer.validate({"data": template_content})


# This needs a strange parametrization. If `unoconv_local` is in a separate
# `parametrize()`, the template filename in the second test will be appended with a
# hash and the test fails
@pytest.mark.parametrize(
    "template__engine,template__template,unoconv_local",
    [
        (models.Template.DOCX_TEMPLATE, django_file("docx-template.docx"), True),
        (models.Template.DOCX_TEMPLATE, django_file("docx-template.docx"), False),
    ],
)
def test_template_merge_as_pdf(
    db, settings, unoconv_local, client, mock_filefield_name_validation, template
):
    url = reverse("template-merge", args=[template.pk])

    response = client.post(
        url, data={"data": {"test": "Test input"}, "convert": "pdf"}, format="json"
    )
    assert response.status_code == status.HTTP_200_OK
    assert response["Content-Type"] == "application/pdf"
    assert f"{template.pk}.pdf" in response["Content-Disposition"]
    assert response.content[0:4] == b"%PDF"


@pytest.mark.parametrize(
    "template__engine,template__template",
    [(models.Template.DOCX_TEMPLATE, django_file("docx-template-loopcontrols.docx"))],
)
def test_template_merge_jinja_extensions_docx(
    db, client, mock_filefield_name_validation, template, settings, snapshot
):
    settings.DOCXTEMPLATE_JINJA_EXTENSIONS = ["jinja2.ext.loopcontrols"]

    url = reverse("template-merge", args=[template.pk])

    response = client.post(url, data={"data": {"test": "Test input"}}, format="json")
    assert response.status_code == status.HTTP_200_OK
    assert (
        response.get("content-type")
        == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    )

    docx = Document(io.BytesIO(response.content))
    xml = etree.tostring(docx._element.body, encoding="unicode", pretty_print=True)
    snapshot.assert_match(xml)


@pytest.mark.parametrize(
    "missing_file,wrong_mime,status_code",
    [
        (False, False, status.HTTP_200_OK),
        (False, True, status.HTTP_400_BAD_REQUEST),
        (True, False, status.HTTP_400_BAD_REQUEST),
    ],
)
@pytest.mark.parametrize(
    "template__engine,template__template",
    [(models.Template.DOCX_TEMPLATE, django_file("docx-template-filters.docx"))],
)
def test_template_merge_jinja_filters_docx(
    db,
    client,
    mock_filefield_name_validation,
    template,
    snapshot,
    settings,
    tmp_path,
    missing_file,
    wrong_mime,
    status_code,
):
    settings.LANGUAGE_CODE = "de-ch"
    url = reverse("template-merge", args=[template.pk])

    # Couldn't put this into `parametrize`. For some reason, in the second run, the
    # template name is extended with a seemingly random string.
    template.template = django_file("docx-template-filters.docx")
    template.save()

    data = {
        "data": json.dumps(
            {
                "test_date": "1984-09-15",
                "test_time": "23:24",
                "test_datetime": "1984-09-15 23:23",
                "test_datetime2": "23:23-1984-09-15",
                "test_none": None,
                "test_nested": {"multiline": "This is\na test."},
            }
        ),
    }

    if not missing_file:
        file = django_file("black.png").file
        if wrong_mime:
            # create a file with the correct filename (black.png) but with
            # the contents of a docx.
            file = tmp_path / "black.png"
            for line in template.template.file:
                file.write_bytes(line)
            file = file.open("rb")

        data["files"] = [file]

    response = client.post(url, data=data, format="multipart")
    assert response.status_code == status_code

    if status_code == status.HTTP_200_OK:
        assert (
            response.get("content-type")
            == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
        )

        docx = Document(io.BytesIO(response.content))
        xml = etree.tostring(docx._element.body, encoding="unicode", pretty_print=True)
        snapshot.assert_match(xml)


@pytest.mark.parametrize(
    "template__engine,template__template",
    [(models.Template.DOCX_TEMPLATE, django_file("docx-template-filters.docx"))],
)
@pytest.mark.parametrize(
    "file_value",
    [None, ""],
)
def test_template_merge_file_reset(
    db,
    client,
    mock_filefield_name_validation,
    template,
    settings,
    file_value,
):
    settings.LANGUAGE_CODE = "de-ch"
    url = reverse("template-merge", args=[template.pk])

    # Couldn't put this into `parametrize`. For some reason, in the second run, the
    # template name is extended with a seemingly random string.
    template.template = django_file("docx-template-filters.docx")
    template.save()

    data = {
        "data": {
            "test_date": "1984-09-15",
            "test_time": "23:24",
            "test_datetime": "1984-09-15 23:23",
            "test_datetime2": "23:23-1984-09-15",
            "test_none": None,
            "test_nested": {"multiline": "This is\na test."},
            "black.png": file_value,
        }
    }

    response = client.post(url, data=data, format="json")
    assert response.status_code == status.HTTP_200_OK

    assert (
        response.get("content-type")
        == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    )


@pytest.mark.parametrize(
    "sample,expected",
    [
        ({"foo": {"bar": ["foo", "blah"]}}, ["foo", "foo.bar", "foo.bar[]"]),
        (
            {
                "this": {
                    "is": {
                        "a": [
                            {
                                "list": {
                                    "with": {
                                        "a": ["nested", "object", "and", "a", "list"]
                                    }
                                }
                            }
                        ]
                    }
                }
            },
            [
                "this",
                "this.is",
                "this.is.a",
                "this.is.a[]",
                "this.is.a[].list",
                "this.is.a[].list.with",
                "this.is.a[].list.with.a",
                "this.is.a[].list.with.a[]",
            ],
        ),
    ],
)
def test_sample_to_placeholders(sample, expected):
    ts = serializers.TemplateSerializer()
    assert ts._sample_to_placeholders(sample) == sorted(expected)


@pytest.mark.parametrize(
    "template__engine,template__template",
    [
        (
            models.Template.DOCX_TEMPLATE,
            django_file("docx-template-placeholdercheck.docx"),
        )
    ],
)
def test_template_merge_missing_data(
    db, client, mock_filefield_name_validation, template, settings
):
    settings.DOCXTEMPLATE_JINJA_EXTENSIONS = ["jinja2.ext.loopcontrols"]

    url = reverse("template-merge", args=[template.pk])

    response = client.post(url, data={"data": {"blah": "Test input"}}, format="json")

    assert response.status_code == status.HTTP_400_BAD_REQUEST
    assert response.json() == [
        "Placeholder from template not found in data: 'bar' is undefined"
    ]



import imghdr

from babel.dates import format_date, format_datetime, format_time
from dateutil.parser import parse
from django.conf import settings
from django.utils.translation import to_locale
from docx.shared import Mm
from docxtpl import InlineImage, Listing
from jinja2 import Environment, pass_context
from rest_framework.exceptions import ValidationError


def parse_string(value):
    return parse(str(value))


def dateformat(value, format="medium", locale=None):
    if value is None:
        return ""

    if locale is None:
        locale = to_locale(settings.LANGUAGE_CODE)

    parsed_value = parse_string(value)
    return format_date(parsed_value, format, locale=locale)


def datetimeformat(value, format="medium", locale=None):
    if value is None:
        return ""

    if locale is None:
        locale = to_locale(settings.LANGUAGE_CODE)

    parsed_value = parse_string(value)
    return format_datetime(parsed_value, format, locale=locale)


def timeformat(value, format="medium", locale=None):
    if value is None:
        return ""

    if locale is None:
        locale = to_locale(settings.LANGUAGE_CODE)

    parsed_value = parse_string(value)
    return format_time(parsed_value, format, locale=locale)


def emptystring(value):
    if value is None:
        return ""
    return value


def getwithdefault(value, default=""):
    if value is None:
        return default
    return value


def multiline(value):
    return Listing(value)


@pass_context
def image(ctx, img_name, width=None, height=None):
    tpl = ctx["_tpl"]

    if img_name not in ctx:
        raise ValidationError(f'No file for image "{img_name}" provided!')

    img = ctx.get(img_name)

    if not img:
        # Fallback to no image
        return

    img.seek(0)  # needed in case image is referenced multiple times
    if imghdr.what(img) not in ["png", "jpg", "jpeg"]:
        raise ValidationError("Only png and jpg images are supported!")

    width = Mm(width) if width else None
    height = Mm(height) if height else None
    return InlineImage(tpl, img, width=width, height=height)


def get_jinja_filters():
    return {
        "date": dateformat,
        "datetime": datetimeformat,
        "time": timeformat,
        "emptystring": emptystring,
        "getwithdefault": getwithdefault,
        "multiline": multiline,
        "image": image,
    }


def get_jinja_env():
    jinja_env = Environment(extensions=settings.DOCXTEMPLATE_JINJA_EXTENSIONS)
    jinja_env.filters.update(get_jinja_filters())
    return jinja_env



import json

from django.db.models.fields.json import KeyTextTransform
from django_filters import Filter, FilterSet
from django_filters.constants import EMPTY_VALUES
from rest_framework.exceptions import ValidationError

from . import models


# TODO: refactor into reusable package later
class JSONValueFilter(Filter):
    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        valid_lookups = self._valid_lookups(qs)

        try:
            value = json.loads(value)
        except json.decoder.JSONDecodeError:
            raise ValidationError("JSONValueFilter value needs to be json encoded.")

        if isinstance(value, dict):
            # be a bit more tolerant
            value = [value]

        for expr in value:
            if expr in EMPTY_VALUES:  # pragma: no cover
                continue
            if not all(("key" in expr, "value" in expr)):
                raise ValidationError(
                    'JSONValueFilter value needs to have a "key" and "value" and an '
                    'optional "lookup" key.'
                )

            lookup_expr = expr.get("lookup", self.lookup_expr)
            if lookup_expr not in valid_lookups:
                raise ValidationError(
                    f'Lookup expression "{lookup_expr}" not allowed for field '
                    f'"{self.field_name}". Valid expressions: '
                    f'{", ".join(valid_lookups.keys())}'
                )
            # "contains" behaves differently on JSONFields as it does on TextFields.
            # That's why we annotate the queryset with the value.
            # Some discussion about it can be found here:
            # https://code.djangoproject.com/ticket/26511
            if isinstance(expr["value"], str):
                qs = qs.annotate(
                    field_val=KeyTextTransform(expr["key"], self.field_name)
                )
                lookup = {f"field_val__{lookup_expr}": expr["value"]}
            else:
                lookup = {
                    f"{self.field_name}__{expr['key']}__{lookup_expr}": expr["value"]
                }
            qs = qs.filter(**lookup)
        return qs

    def _valid_lookups(self, qs):
        # We need some traversal magic in case field name is a related lookup
        traversals = self.field_name.split("__")
        actual_field = traversals.pop()

        model = qs.model
        for field in traversals:  # pragma: no cover
            model = model._meta.get_field(field).related_model

        return model._meta.get_field(actual_field).get_lookups()


class TemplateFilterSet(FilterSet):
    meta = JSONValueFilter(field_name="meta")

    class Meta:
        model = models.Template
        fields = {
            "slug": ["exact"],
            "description": ["icontains", "search"],
        }



from factory import Faker
from factory.django import DjangoModelFactory

from . import models


class TemplateFactory(DjangoModelFactory):
    slug = Faker("slug")  # type: ignore
    description = Faker("text")
    engine = Faker("word", ext_word_list=models.Template.ENGINE_CHOICES_LIST)
    template = None
    meta = {}  # type: ignore

    class Meta:
        model = models.Template



import pytest

from ..jinja import dateformat, datetimeformat, emptystring, getwithdefault, timeformat


@pytest.mark.parametrize(
    "inp,locale,expected", [("2019-12-31", "de_CH", "31.12.2019"), (None, "de_CH", "")]
)
def test_dateformat(inp, expected, locale):
    formatted = dateformat(inp, locale=locale)
    assert formatted == expected


@pytest.mark.parametrize(
    "inp,locale,expected",
    [("2019-12-31 23:59", "de_CH", "31.12.2019, 23:59:00"), (None, "de_CH", "")],
)
def test_datetimeformat(inp, expected, locale):
    formatted = datetimeformat(inp, locale=locale)
    assert formatted == expected


@pytest.mark.parametrize(
    "inp,locale,expected", [("23:59", "de_CH", "23:59:00"), (None, "de_CH", "")]
)
def test_timeformat(inp, expected, locale):
    formatted = timeformat(inp, locale=locale)
    assert formatted == expected


@pytest.mark.parametrize("inp,expected", [("text", "text"), (None, "")])
def test_emptystring(inp, expected):
    formatted = emptystring(inp)
    assert formatted == expected


@pytest.mark.parametrize(
    "inp,default,expected",
    [("text", "", "text"), (None, "", ""), (None, "something", "something")],
)
def test_getwithdefault(inp, default, expected):
    formatted = getwithdefault(inp, default=default)
    assert formatted == expected



import os.path
import shutil

from django.conf import settings
from django.core.files import File

_data_path = os.path.dirname(os.path.realpath(__file__))


def django_file(name, mode="rb"):
    abspath = os.path.join(_data_path, name)
    new_path = f"{settings.MEDIA_ROOT}/attachments"

    try:
        os.makedirs(new_path)
    except FileExistsError:  # pragma: no cover
        pass

    shutil.copy(abspath, f"{new_path}/{name}")

    return File(open(abspath, mode), name=f"attachments/{name}")



from django.apps import AppConfig
from django.conf import settings
from django.db.models import TextField
from django.db.models.lookups import IContains


class DefaultConfig(AppConfig):
    name = "document_merge_service.api"

    def ready(self):
        if "sqlite3" in settings.DATABASES["default"]["ENGINE"]:  # pragma: no cover
            TextField.register_lookup(IContains, lookup_name="search")



# Generated by Django 3.2.9 on 2022-03-29 07:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("api", "0004_cleanup_files"),
    ]

    operations = [
        migrations.AlterField(
            model_name="template",
            name="engine",
            field=models.CharField(
                choices=[
                    (
                        "docx-template",
                        "Docx Template engine (https://github.com/elapouya/python-docx-template)",
                    ),
                    (
                        "docx-mailmerge",
                        "Docx MailMerge engine (https://github.com/Bouke/docx-mailmerge)",
                    ),
                    (
                        "xlsx-template",
                        "Xlsx Template engine (https://github.com/zhangyu836/python-xlsx-template)",
                    ),
                ],
                max_length=20,
            ),
        ),
    ]



from django.conf import settings
from django.conf.urls import include
from django.urls import re_path

urlpatterns = [
    re_path(
        f"^{settings.URL_PREFIX}api/v1/", include("document_merge_service.api.urls")
    )
]



import shutil
from io import BytesIO
from pathlib import Path

import pytest
import sentry_sdk
from django.core.cache import cache
from pytest_factoryboy import register
from rest_framework.test import APIClient

from document_merge_service.api import engines, factories, models
from document_merge_service.api.data import django_file

from .api.authentication import AnonymousUser
from .sentry import sentry_init

register(factories.TemplateFactory)


class TestUser(AnonymousUser):
    def __init__(self, username=None, groups=None):
        self.username = username if username else "admin"
        self.groups = groups or []

    @property
    def is_authenticated(self):
        return True


@pytest.fixture(scope="function", autouse=True)
def _autoclear_cache():
    cache.clear()


@pytest.fixture
def admin_groups():
    return ["admin"]


@pytest.fixture
def admin_user(admin_groups):
    return TestUser(groups=admin_groups)


@pytest.fixture
def client():
    return APIClient()


@pytest.fixture
def admin_client(db, admin_user):
    client = APIClient()
    client.force_authenticate(user=admin_user)
    return client


@pytest.fixture
def mock_filefield_name_validation(mocker):
    mocker.patch(
        "django.db.models.fields.files.validate_file_name",
        side_effect=lambda name, *args, **kwargs: name,
    )


@pytest.fixture
def docx_template_with_placeholder(admin_client, template):
    """Return a factory function to build a docx template with a given placeholder."""
    template.engine = models.Template.DOCX_TEMPLATE
    template.template = django_file("docx-template.docx")
    template.save()

    def make_template(placeholder):
        engine = engines.get_engine(template.engine, template.template)
        binary = BytesIO()
        engine.merge({"test": placeholder}, binary)
        binary.seek(0)
        template.template.save("foo.docx", binary)
        template.save()
        return template

    return make_template


@pytest.fixture
def dms_test_bin():
    sleep_path = Path(shutil.which("sleep"))
    test_path = Path(Path(__file__).parent.absolute(), "tmpb5nw53v5")
    with test_path.open("wb") as f, open(sleep_path, "rb") as g:
        f.write(g.read())
    test_path.chmod(0o755)
    yield test_path
    test_path.unlink()


@pytest.fixture
def loadtest_data():
    base = Path(__file__).parent.absolute()
    return Path(base, "api", "data", "loadtest")


@pytest.fixture
def sentry_mock(monkeypatch, mocker):
    sentry_init("https://SomePublicKey@0.ingest.sentry.io/0", "test", 0.01, False)
    sentry_client = sentry_sdk.Hub.current.client
    transport = mocker.MagicMock()
    monkeypatch.setattr(sentry_client, "transport", transport)
    try:
        yield transport
    finally:
        sentry_sdk.Hub.current.bind_client(None)
